;* 
;* File:         med_copy.s
;* RCS:          $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel/hpsgc/medcopy.s,v 1.4 1994/11/10 17:59:13 mike Exp $
;* Description:  Copy a block of bytes between memory and IO space.
;* Author:       Chris Dalton, High-performance Networking Group, HPL Bristol
;* Language:     PA-RISC 1.1
;*
;*  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
;*
;*  To anyone who acknowledges that this file is provided "AS IS"
;*  without any express or implied warranty:
;*      permission to use, copy, modify, and distribute this file
;*  for any purpose is hereby granted without fee, provided that
;*  the above copyright notice and this notice appears in all
;*  copies, and that the name of Hewlett-Packard Company not be
;*  used in advertising or publicity pertaining to distribution
;*  of the software without specific, written prior permission.
;*  Hewlett-Packard Company makes no representations about the
;*  suitability of this software for any purpose.
;*
;*                         --------------------------
;*
;* MED_COPY
;*
;* unsigned long med_copy(unsigned char *src, unsigned char *dest, unsigned len)
;*
;* Med_copy copies `len' bytes from the contiguous area pointed to by `src'
;* into the area pointed to by `dest'.  If the two areas overlap, the result
;* is undefined.  Its implementation conforms to the following rules.
;*
;*      All accesses to memory must be properly-aligned 32-bit words.
;*      Only words containing bytes affected by the copy are accessed.
;*      Bytes outside the range of the copy are preserved.
;*
;* Med_copy is necessary because memcpy and bcopy may violate the first rule.
;*
;* Med_copy returns the 32-bit one's complement sum of the data moved to
;* or from the card.  See RETURN VALUE below.
;*
;*                         --------------------------
;*
;* MED_VCOPY
;*
;* unsigned long med_vcopy(unsigned char *src, unsigned char *dest, unsigned len)
;*
;* Med_vcopy is identical in effect to med_copy, but uses an implementation
;* technique which copies data from IO space to memory nearly twice as
;* quickly as med_copy.  However, its use is subject to a couple of
;* limitations:
;*
;*      Use of Venom:  the caller must have the right privilege and mapping
;*      to write to Venom in IO space, otherwise a data protection trap or
;*      data page fault will occur.  Also note that the present version does
;*      not preserve Venom state; this may have adverse effects on certain
;*      graphics programs.
;*
;*      Amount of data:  currently, med_vcopy will only work for IO reads
;*      of amounts up to 16K bytes.  It does not fail gracefully.
;* 
;* Med_vcopy returns the 32-bit one's complement sum of the data moved to
;* or from the card.  See RETURN VALUE below.
;*
;*
;* PERFORMANCE
;*
;* Med_copy and med_vcopy are designed to give (nearly) the best possible
;* throughput for copying between memory (cache) and IO space (ie, Medusa
;* buffers).  They copy from memory to memory at very nearly the same rate
;* as memcpy (50 MB/s on a 720), write to IO from cache at 35 to 38 MB/s,
;* and read from IO to cache at 10 MB/s (med_copy) or 18.5 MB/s (med_vcopy).
;*
;*
;* WARNINGS
;*
;* Many branch instructions have something useful happening in the delay
;* slot.  Don't insert code willy-nilly.
;*
;* Med_vcopy only works properly for reads of up to 16 Kbytes.  This is more
;* than adequate for medusa, so we have left it for the time being.
;*
;* If you're using med_copy or med_vcopy to write to Medusa or Afterburner
;* using the built-in checksum unit, remember that the functions READ from
;* the destination when either `dest' or `len' is not a multiple of four
;* (ie, the copy is not dealing with whole words), and that this will affect
;* the result.
;*
;*
;* RETURN VALUE (Checksum)
;*
;* If either `src' or `dest' is an address in the checksummed area of
;* the Medusa card, med_copy or med_vcopy uses Medusa's built-in checksum
;* unit to calculate a value which may be used to calculate the Internet
;* checksum of the data copied.
;*
;* The checksum is the 32-bit one's-complement sum of the smallest set of
;* 32-bit words that contains the `len' bytes of data in the Medusa buffer.
;* Where the start or end of the data do not correspond with a word boundary,
;* the calculation behaves as if the data were padded with the appropriate
;* number of zeros.
;* 
;*
;* DESIGN and IMPLEMENTATION NOTES
;*
;* Some notes about med_copy and med_vcopy appear at the end of this file.
;*
;*
;* ASSEMBLING THE FILE
;*
;* You must apply the C pre-processor to this file before you try to
;* assemble it with 'as'.  Using 'cc' or 'c89' does this for you.
;*
;* Defining the following symbols will affect the generated code.  Neither
;* is defined by default.
;*
;*      TESTING - alters names of entry points, to allow comparitive testing
;*                with other versions of med_copy.
;*          defined:      entry points are `med_copy_t' and `med_vcopy_t'
;*          not defined:  entry points are `med_copy' and `med_vcopy'
;*
;*      RETURN_PTRS - when defined, alters med_copy to return the final
;*          values of src and dest.  Dest is always the address of the
;*          word following the last word affected by the copy.  Src is
;*          the address of the word after the last word of the source area
;*          for med_copy, but is undefined for med_vcopy.  To retrieve the
;*          values in a program, declare med_copy as
;*
;*            struct med_copy_result{ caddr_t src; caddr_t dest; };
;*            struct med_copy_result med_copy();
;*
;*      ALWAYS_LOAD_CACHE - controls whether aligned venom reads bring the
;*          data into the cache.  If defined, they do (with some loss of
;*          throughput); if not, they don't.  Non-aligned reads manipulate
;*          the data, so always bring it into cache.


#include <medusa.h>
#if NMEDUSA > 0

        .version "@(#) med_vcopy.s $Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel/hpsgc/medcopy.s,v 1.4 1994/11/10 17:59:13 mike Exp $"

        .space   $TEXT$
        .subspa  $CODE$
#ifdef TESTING
        .export  med_copy_t
        .export  med_vcopy_t
#else
        .export  med_copy
        .export  med_vcopy
#endif


; STACK FRAME =============================================
;
;  SP      -----------------
;          Frame Marker Area
;  SP-32   -----------------
;          Fixed Arguments Area
;            flags_s - control flags (io_read, io_write, use_venom, do_cksum)
;            cksum_unit - address of checksum unit for the card
;  SP-48   -----------------
;          Variable Arguments Area
;            stage_buf - staging area for data during shifted Venom moves
;  SP-144  -----------------

; CONSTANTS ===============================================
;
; Stack frame
;
;   Area sizes

sz_stage_buf    .equ 96
sz_std_frame    .equ 32
sz_std_args     .equ 16

;   Offsets from stack pointer, eg, for `ldw xxx(sp), rx'

stage_buf       .equ  -(sz_std_frame+sz_std_args+sz_stage_buf)      ; -144
flags_s         .equ  -(sz_std_frame+sz_std_args)                   ; -48
cksum_unit      .equ  -(sz_std_frame+sz_std_args)+4                 ; -44
spare2          .equ  -(sz_std_frame+sz_std_args)+8                 ; -40
spare3          .equ  -(sz_std_frame+sz_std_args)+12                ; -36


; Medusa constants

;   Checksummed VRAM

cksum_base  .equ  0xf3800000  ; (addr & cksum_base == cksum_base) => checksum
cksum_bit   .equ  8           ; in medusa address, 1 => checksum, 0 => not.

;   Hardware registers

card_base_mask  .equ  0xfc000000
cksum_unit_off  .equ  0x01400000


; Flags in reg `flags'  (Use _b variant in bit set/clear/test instructions)

use_venom     .equ   1   ; set iff copy is suitable for venom read.
use_venom_b   .equ  31
io_write      .equ   2   ; set iff copy is write to IO space
io_write_b    .equ  30
io_read       .equ   4   ; set iff copy is read from IO space.
io_read_b     .equ  29
do_cksum      .equ   8   ; set iff src or destination is in IO space
do_cksum_b    .equ  28

; Venom control register addresses

venom_user_base       .equ  0xfffbc080
venom_user_ctrl       .equ  0
venom_span_count      .equ  4
venom_graphics_addr   .equ  8

venom_supervisor_ctrl .equ  0xfffbd080
sgcslot_baseaddr_mask .equ  0x0c000000


; REGISTER ASSIGNMENTS ====================================

; Name   Assignment     ; conventional name (where given name is valid)
                        ; What it is

src     .reg  %r26      ; arg0 (everywhere)
                        ; Address of first byte/word of source data.
dest    .reg  %r25      ; arg1 (everywhere)
                        ; Address of first byte/word of destination area.
length     .reg  %r24      ; arg2 (initially)
                        ; Number of bytes to copy

ntw     .reg  %r24      ; arg2 (after ~10 instructions)
                        ; Number of bytes left to write to destination.
buf0    .reg  %r28      ; ret0 (during venom copy loops)
                        ; Address of "staging buffer" for incoming data.
flags   .reg  %r29      ; ret1 (prefix and copy loops before use of Venom)
                        ; Control flags (see definitions above)
venom   .reg  %r29      ; ret1 (before main copy loops)
                        ; Initially, 1 => use venom, 0 => don't.
                        ; Later, holds address of venom control registers.
dest_nock .reg %r2      ; rp   (in prefix and suffix)
                        ; ==nt to dest except not in checksummed space.
link    .reg  %r2       ; rp   (in main body)
                        ; Return address for `setup_venom' subroutine.

ntr     .reg  %r23      ; arg3 (after ~10 instructions)
                        ; Number of bytes left to read from source.

shiftb   .reg  %r31      ; (before copy loops)
                        ; Nr. of bits to shift data `right' during copy.
chunk   .reg  %r1       ; (any time)
                        ; Scratch - usually holds the nr of bits or bytes
                        ; involved in some operation
wk1     .reg  %r19      ; (any time)
                        ; Conveys unprocessed data between copy stages
wk2     .reg  %r20      ; (any time)
                        ; Sometimes contains data from destination for merging
wk3     .reg  %r21      ; (during copy loops)
wk4     .reg  %r22      ; (during copy loops)
wk5     .reg  %r31      ; (during copy loops)
                        ; Hold data in transit during copy loops.

delta_s .reg  %r21      ; (before copy loops)
                        ; Byte offset of *src within word (== src % 4)
delta_d .reg  %r22      ; (before copy loops)
                        ; Byte offset of *dest within word (== dest % 4)
sp	.reg  %r30	; stack pointer
ret0	.reg  %r28	; return register

; CODE ====================================================

        .code

        .proc
        .callinfo  caller,frame=96,save_rp
#ifdef TESTING
med_copy_t
#else
med_copy                                  ; ENTRY POINT - med_copy
#endif
        movb,tr,n 0, flags, entry
#ifdef TESTING
med_vcopy_t
#else
med_vcopy                                 ; ENTRY POINT - med_vcopy
#endif
        ldi       use_venom, flags
entry
        .enter

; if (length == 0) return;
        comb,=,n  0, length, return

; //  Set flags appropriate to src address:
; //    io_read - reading from IO space  ( src == 0xf....... )
; //    use_venom - io_read && entered at med_vcopy.
; //    do_cksum - io_read and src in checksummed address range (0xf.8.....)

; wk1= 0;
; if (src in IO space)
        zdepi     0xf, 3, 4, chunk
        comb,<<   src, chunk, src_checked
        copy      0, wk1
;   {
;     io_read= TRUE;
;     if (src is in checksummed medusa area)
          ldil    l%cksum_base, wk2   
          and     src, wk2, chunk
          comb,<> wk2, chunk, src_checked
	  depi    1, io_read_b, 1, flags
;       {
;         do_cksum= TRUE;
            depi  1, do_cksum_b, 1, flags
;         wk1= src;          /* for compensation calculation */
            copy  src, wk1
;       }
;   }
src_checked

; |- io_read == (src is in IO space)
; |- use_venom == io_read && called as med_vcopy
; |- do_cksum == io_read && (src is in checksummed VRAM)
; |- wk1 == (do_cksum) ? src : 0

; delta_s= src % sizeof(unsigned long);
        extru   src, 31, 2, delta_s
;     src-= delta_s
        depi    0, 31, 2, src


; //  Set flags appropriate to dest address.  Note that the needs of the
; //  destination have priority over those of the source.
; //    io_write == writing to IO space  (dest == 0xf....... )
; //    use_venom == io_write ? FALSE : (io_read && entered at med_vcopy)
; //    do_cksum == io_write && dest in checksummed address range (0xf.8.....)

; dest_nock= dest;
; if (dest in IO space)
        zdepi     0xf, 3, 4, chunk
        comb,<<   dest, chunk, dest_checked
        copy      dest, dest_nock
;   {
;     use_venom= FALSE;
        depi      0, use_venom_b, 1, flags
;     io_write= TRUE;
        depi      1, io_write_b, 1, flags

;     if (dest is in checksummed medusa area)
          ldil    l%cksum_base, wk2   
          and     dest, wk2, chunk
          comb,<>,n wk2, chunk, dest_checked
;       {
;         if (wk1 != 0)  // src is also in cksum area, so alter it so it isn't.
              extru,= wk1, 31, 32, 0
;           {
;             src&= ~0x00800000;
                dep 0, cksum_bit, 1, src
;           }

;         do_cksum= TRUE;
            depi  1, do_cksum_b, 1, flags
;         dest_nock&= ~0x00800000; 
            dep   0, cksum_bit, 1, dest_nock
;         wk1= dest;         /* for compensation calculation */
            copy  dest, wk1
;       }
;   }
dest_checked

; |- io_read == (src in IO space)
; |- io_write == (dest in IO space)
; |- do_cksum == (src or dest in checksummed VRAM)
; |- use_venom == !io_write && io_read && (called as med_vcopy)
; |- wk1 == (writing to checksummed VRAM) -> (checksumming dest)
;            (reading checksummed VRAM) -> (checksumming src)
;            (otherwise) -> 0
; |- dest_nock == (dest in checksummed VRAM) ? (non-checksumming dest) : dest
; |- src' == (src and dest in checksummed VRAM) ? (non-checksumming src) : src

; flags_s= flags
        stw     flags, flags_s(sp)

; delta_d= dest % sizeof(unsigned long);
        extru   dest, 31, 2, delta_d
; dest-= delta_d;
        dep     0, 31, 2, dest
; dest_nock-= delta_d;
        dep     0, 31, 2, dest_nock

; // Calculate the initial value for the checksum unit.  This is the one's
; // complement of the one's complement sum of the parts of the first and
; // last words in the copy range which are NOT involved in the copy.
; // Set cksum_unit to the address of the checksum unit.

; cksum_unit= 0;
; if (do_cksum)
      bb,>=   flags, do_cksum_b, done_cksum_comp
      stw     0, cksum_unit(sp)
;   {
;     |- wk1 is byte address of area in checksum RAM (could be src or dest)
;     |- length is number of bytes in area.

;     cksum_unit= address of checksum unit corresponding to address in wk1.
        ldil  l%card_base_mask, wk2
        and   wk1, wk2, wk2
        addil l%cksum_unit_off, wk2 ; sum --> chunk (%r1)
        stw   chunk, cksum_unit(sp)

;     wk1= non-checksumming version of wk1;
        dep   0, cksum_bit, 1, wk1

;     // Extract uncopied bytes at start of area into wk2.

;     wk2= 0;
;     if (chunk= wk1 & 3)
          zdep,<>       wk1, 31, 2, chunk
          b     ckcomp_end
          copy  0, wk2
;       {
;         buf0= address of word containing *wk1.
            sub     wk1, chunk, buf0
;         wk2= bytes at *buf0 not affected by the copy.
            ldws    (buf0), wk2
            ldo     -1, shiftb
            sh3add  chunk, shiftb, shiftb
            mtsar   shiftb
            vextru  wk2, 31, wk2
            zvdep   wk2, 31, wk2
;       }
ckcomp_end

;     // Extract uncopied bytes at end of area into wk1
;     // and form one's complement sum of wk1 and wk2 in wk2.
;     wk1+= length;
            add     length, wk1, wk1
;     if (chunk= wk1 & 3)
          zdep,<>       wk1, 31, 2, chunk
          b,n   ckcomp_save
;       {
;         buf0= address of word containing *wk1.
            sub     wk1, chunk, buf0
;         wk1= bytes at *buf0 not affected by the copy.
            ldws    (buf0), wk1
            ldo     -1, shiftb
            sh3add  chunk, shiftb, shiftb
            mtsar   shiftb
            vdep    0, 31, wk1

;         wk2= one's complement sum of wk1 and wk2
            add     wk1, wk2, wk2
            addc    0, wk2, wk2
            addc    0, wk2, wk2
;       }
ckcomp_save

;     *cksum_unit= ~wk2;
        ldw     cksum_unit(sp), wk1
        uaddcm  0, wk2, wk2
        stw     wk2, (wk1)

;   }
done_cksum_comp


; ntr= length+delta_s;
        add     length, delta_s, ntr
; ntw= length+delta_d;
        add     length, delta_d, ntw

; wk1= *(src++);  ntr-= 4;
        ldo     -4(ntr), ntr    ; ldws deferred to following branch delay slot

; if (0 == (shiftb= 8*(delta_d - delta_s))
        sub     delta_d, delta_s, shiftb
        zdep,=  shiftb, 28, 31, shiftb
        b       n_prefix
        ldws,ma 4(0,src), wk1
; {

a_prefix  ; pronounced "Aligned prefix"

;   |-   src -> word containing first byte of source
;   |-   dest -> word containing first byte of destination
;   |-   ntw == original length + delta_d
;   |-   ntr == original length + delta_s
;   |-   delta_d == nr of bytes to left of first byte of src/dest.

;   if (0 < (chunk= 8*delta_d - 1))
      sh3add        delta_d, 0, chunk
      addibt,<,n  -1, chunk, a_prefix_done
;     {
;       wk2= *d;
          ldws      (0,dest_nock), wk2
          mtsar     chunk

;       wk1= combine(wk1, wk2);
          vextru        wk2, 32, wk2
          vdep  wk2, 32, wk1
;     }
a_prefix_done

;       if (ntw <= 4) goto suffix;
          comib,>=,n  4, ntw, suffix

;       *(d++)= wk1;
          stws,ma wk1, 4(0,dest)
          ldo   -4(ntw), ntw

; = ALIGNED COPY ==========================================

;   |- ntw == number of bytes to copy from *src to *dest
;   |- ntr not set.

a_copy
;   if (! use_venom)
        bb,<,n      flags, use_venom_b, vr_a_copy
;     {
;       if (! io_write)
#ifdef TEST_WRITE
            b,n pw_a_copy
#else
            bb,<,n  flags, io_write_b, pw_a_copy
#endif
pr_a_copy                               ; Programmed Read, aligned copy
;         while (ntw >= 16)             ;  (copy 16 bytes at a time)
            comib,>=,n  15, ntw, pw_a_copy
pr_a_loop                               
;           {
                ldws,ma   4(0,src), wk1
                ldws,ma   4(0,src), wk2
                ldws,ma   4(0,src), wk3
                ldws,ma   4(0,src), wk4

                stws,ma   wk1, 4(0,dest)
                stws,ma   wk2, 4(0,dest)
                ldo       -16(ntw), ntw
                stws,ma   wk3, 4(0,dest)
                comib,<   15, ntw, pr_a_loop
                stws,ma   wk4, 4(0,dest)
;           }
pw_a_copy                               ; Programmed Write, aligned copy
;       while (ntw >= 8)                ;   (copy 8 bytes at a time)
          comib,>=,n  7, ntw, p_a_tail
pw_a_loop
;         {
              ldws,ma   4(0,src), wk1
              ldws,ma   4(0,src), wk2
              stws,ma   wk1, 4(0,dest)
              ldo       -8(ntw), ntw
              comib,<   7, ntw, pw_a_loop
              stws,ma   wk2, 4(0,dest)
;         }

p_a_tail
;       while (ntw > 0)
            comb,>=,n    0, ntw, p_a_copy_done
;         {
p_a_tail_loop
            ldws,ma   4(0,src), wk1
            comib,>=,n  4, ntw, p_a_copy_done
            stws,ma   wk1, 4(0,dest)
            b         p_a_tail_loop
            ldo       -4(ntw), ntw
;         }
p_a_copy_done
        b,n suffix
;     }
;   else
;     {
vr_a_copy                               ; Venom Read, aligned copy

;       if (ntw > 4)
          comibf,<=,n   4, ntw, vr_a_last_word
;         {
;           |-  src -> first word of the rest of the copy
;           |-  nr of words to copy = ntw/4 - 1 (ie, read all whole words with
;               venom, and check for 1 to 3 straggling bytes at the end)

;           setup_venom(block_move, ntw/4 - 1, s)
              extru   ntw, 29, 29, chunk
              bl      setup_venom, link
              ldo     -1(chunk), chunk


;           // 1st block - copy at most 32 bytes to make dest cache-aligned.

;           chunk= nr of bytes in first cache line not affected by venom move.

;           if (chunk != 0)
              extru,<>  dest, 31, 5, chunk
              b,n       vr_a_second_chunk
;             {
;               chunk= nr of bytes in first cache line AFFECTED by venom move
                  subi    32, chunk, chunk

;               if (ntw < chunk) chunk= ntw;
                  sub,>=  ntw, chunk, 0
                  copy    ntw, chunk
                  dep     0, 31, 2, chunk

;               venom_read(dest, chunk);  dest+= chunk;
                  gfw,m   chunk(dest)

;               ntw-= chunk;
                  sub     ntw, chunk, ntw
;             }

vr_a_second_chunk

;           |- ntw == nr of bytes to write
;           |- dest == addr of next word to read  && dest%32 == 0
;           |- ntr ==  ??

;           chunk= 32;
              addi,tr 32, 0, chunk  ; WARNING:  skips next instruction (gfw,m)

;           while (ntw >= chunk)
;             {
;               ntw-= chunk;
;               venom_read(dest, chunk);  dest+= chunk;

vr_a_venom_loop
                  gfw,m     chunk(dest)
#ifdef ALWAYS_LOAD_CACHE
                  ldw       -32(dest), 0    ; CACHE HINT
#endif
                  comb,>=,n ntw, chunk, vr_a_venom_loop
                  sub       ntw, chunk, ntw
;             }

vr_a_last_chunk

;           |- ntw < 32
;           |- ntw = nr of bytes left in venom_read

;           if (chunk= ntw - ntw%4)  /* chunk= 4*nr of whole words left in venom*/
              copy      ntw, chunk
              dep,<>    0, 31,2, chunk
              b,n       vr_a_last_word
;             {
;               venom_read(dest, chunk);  dest+= chunk;
                  gfw,m   chunk(dest)
;               ntw-= chunk;
                  sub       ntw, chunk, ntw
;             }

;         }

;       |- dest -> last word affected by move
;       |- 0 <= ntw < 4 && ntw == nr of bytes to copy to last word of dest

vr_a_last_word

;       if (ntw)
          comb,=        ntw, 0, return
;         {

;           wk1= src[ntr/4];
        add   src, ntr, src
        dep   0, 31, 2, src
        ldws,ma 4(0,src), wk1

;           ntr%= 4;
;             extru ntr, 31, 2, ntr

;           ntr-= 4;   /* preserve  src_0 + length == src + ntr relationship */
;             ldo   -4(ntr), ntr

;           ntw-= 4;   /* to meet published requirements of suffix */
              ldo   -4(ntw), ntw
;         }

;     }
        b,n     suffix
; }

; else == NON-ALIGNED MOVES ===============================

; {

n_prefix  ; pronounced "Nonaligned prefix"

;   |-   src -> word containing first byte of source
;   |-   dest -> word containing first byte of destination
;   |-   ntw == original length + delta_d
;   |-   ntr == original length + delta_d 
;   |-   delta_d == nr of bytes to left of first byte of dest.
;   |-   delta_s == nr of bytes to left of first byte of src.
;   |-   shiftb == 8 * (delta_d - delta_s)

;   |-  wk1 = [xxxx0011]

;   if (delta_d < delta_s)   /* shifting left */

        combf,<   delta_d, delta_s, n_right_shift
        ldi       -1, chunk

;     {
;       |- chunk == -1
;       |- delta_d < delta_s

;       if ((chunk+= 8*delta_d) > 0) /* have prefix to contend with */
          sh3add,>  delta_d, chunk, chunk
          b,n       n_l_no_prefix

;         {

;           wk2= *dest;
              ldws  0(0, dest_nock), wk2

;           |- wk2.wk1 = [xxAABBCC][xxxx0011]

;           wk2>>= (8*(4-delta_d));

              mtsar chunk
              vextru wk2, 32, wk2

;           |- wk2 = [xxxxxxAA][xxxx0011]

              sub   chunk, shiftb, chunk         ; chunk= delta_s - 1
              mtsar chunk
              vdep  wk2, 32, wk1

;           |- wk1 = [xxAA0011]
;         }

n_l_no_prefix

;       |- wk1 == eventual contents of m.s. end of *dest., aligned with source

;       shift_amount= shiftb;
;       if (ntr <= 0) goto n_fixup;
            comb,>= 0, ntr, n_fixup
            mtsar shiftb
n_l_more
;     }
        b,n   n_copy
;   else /* shifting right */
n_right_shift
;     {
;       |- chunk == -1
;       |- delta_d > delta_s
;       |- wk1 == first word of source 
;       |-  wk1 = [00112233]  //  [xx001122]

;       wk2= *dest >> 8*(4-delta_d);
          ldws  (0,dest_nock), wk2

;       |- wk2 = [AABBxxxx]

          sh3add delta_d, chunk, chunk
          mtsar chunk
          vextru wk2, 32, wk2

;       |- wk2 = [xxxxAABB]
;       |- wk2.wk1 == [xxxxAABB][00112233]  //  [xxxxAABB][xx001122]

;       if (delta_s)
          comb,=,n  0, delta_s, n_r_got_prefix

;         {
;           |- wk2.wk1 == [xxxxAABB][xx001122]

;           wk1{0..8*delta_s-1}= wk2{32-delta_s..31};
              sub   chunk, shiftb, chunk
              mtsar chunk
              vdep  wk2, 32, wk1

;           |- wk2.wk1 ==  [xxxxAABB][BB001122]

;           wk2>>= 8*delta_s;
              ldo   1(chunk), chunk
              mtsar chunk
              vshd  wk2, wk2, wk2

;           |- wk2.wk1 ==  [xxxxxxAA][BB001122]

;         }
n_r_got_prefix
;       |- wk2.wk1 ==  [xxxxAABB][00112233]  //  [xxxxxxAA][BB001122]

;       shift_amount= shiftb;
;       if (ntw <= 4)
            comib,< 4, ntw, n_r_more
            mtsar   shiftb
;         {
;           wk1= [wk2.wk1] >> shiftb;
;           goto suffix
              b     suffix
              vshd  wk2, wk1, wk1
;         }
;       else
n_r_more
;         {
;           wk2= [wk2.wk1] >> shiftb;
              vshd  wk2, wk1, wk2
;           *(dest++)= wk2;  ntw-= 4;
              stws,ma   wk2, 4(0,dest)
              ldo       -4(ntw), ntw
;         }
n_r_done
;     }

n_copy

;   shift_amount= shiftb;
;   if (! use_venom)
        bb,< flags, use_venom_b, vr_n_copy
        mtsar   shiftb
;     {
;       if (! io_write)
#ifdef TEST_WRITE
            b,n pw_n_copy
#else
            bb,<,n      flags, io_write_b, pw_n_copy
#endif
pr_n_copy                               ; Programmed read, nonaligned copy
;         while (ntw >= 16)             ;   (copy 16 bytes at a time)
            comib,>=,n  15, ntw, pw_n_copy
pr_n_loop
;           {
              ldws,ma   4(0,src), wk2
              ldws,ma   4(0,src), wk3
              ldws,ma   4(0,src), wk4
              ldws,ma   4(0,src), wk5

              vshd      wk1, wk2, wk1
              stws,ma   wk1, 4(0,dest)
              ldo       -16(ntr), ntr
              vshd      wk2, wk3, wk2
              stws,ma   wk2, 4(0,dest)
              ldo       -16(ntw), ntw
              vshd      wk3, wk4, wk3
              stws,ma   wk3, 4(0,dest)
              vshd      wk4, wk5, wk4
              stws,ma   wk4, 4(0,dest)
              comib,<    15, ntw, pr_n_loop
              copy      wk5, wk1
;           }
pw_n_copy                               ; Programmed write, nonaligned copy
;       while (ntw >= 8)                ;   (copy 8 bytes at a time)
          addib,<,n   -8, ntw, p_n_loop_end
pw_n_loop
;         {
            ldws,ma   4(0,src), wk2
            ldws,ma   4(0,src), wk3

            vshd      wk1, wk2, wk1
            stws,ma   wk1, 4(0,dest)
            vshd      wk2, wk3, wk2
            ldo       -8(ntr), ntr
            stws,ma   wk2, 4(0,dest)
            addibf,<  -8, ntw, pw_n_loop
            copy      wk3, wk1
;         }
p_n_loop_end
            ldo 8(ntw), ntw
p_n_tail                                ; Programmed IO, nonaligned copy
            b,n  n_tail_entry
;     }
;   else
;     {

vr_n_copy                               ; Venom Read, non-aligned copy

;       |- wk1 contains eventual m.s. end of *dest, aligned as source.
;       |- dest + ntw == original dest + original len
;       |- dest -> next word of destination to receive data.
;       |- src + nt  == original src + original len
;       |- src -> next word of source to be read.
;       |- shiftb == 8 * (delta_d - delta_s)
;       |- 0 < ntw - ntr < 4


;       if (ntr > 0)
;         {

;           find buffer and ensure it's cache-line aligned.

;           buf0= (&<buffer on stack> + 31) & -32
;           // The staging buffer area is 96 bytes starting at%sp-144.
;           // (144 is 96 + 48; 48 is the size of the basic frame)

              ldo   -144+31(%r30), buf0
              dep   0, 31, 5, buf0

;           setup_venom(block_move, (ntr-1)/4, s)
              ldo   -1(ntr), chunk
              bl    setup_venom, link
              extru chunk, 29, 30, chunk

;           venom_read(buf0);   // start loading staging buffer from IO space
              gfw   (buf0)
              ldws  (buf0), 0  ;// CACHE HINT - try to load staging buffer

;           |- wk1 contains eventual m.s. end of *dest, aligned as source.
;           |- ntw == nr of bytes to write
;           |- dest == addr of next word to read  && dest%32 == 0

;           chunk= 32;

              ldo   -32(ntw), ntw
              b   vr_n_venom_entry
              ldi 32, chunk

;           while (ntw >= chunk)
;             // ntw because all 32 bytes the loop body writes are significant,
;             // but the last few (< 4) bytes it reads may not be
;             {
;               ntw-= chunk;
;               ntr-= chunk;

vr_n_venom_loop

                  ldo     -32(ntr), ntr
;               if (ntr > 0) {load other staging buffer from IO space}
                  sub,<=  ntr, 0, 0
                  gfw     chunk,(buf0)

;               copy & shift 32 bytes from wk1 and current staging buffer to *dest

                  ldw     0(buf0), wk2
                  ldw     4(buf0), wk3
                  ldw     8(buf0), wk4
                  ldw     12(buf0), wk5

                  vshd    wk1, wk2, wk1
                  stws,ma wk1, 4(dest)
                  vshd    wk2, wk3, wk2
                  stws,ma wk2, 4(dest)
                  vshd    wk3, wk4, wk3
                  stws,ma wk3, 4(dest)
                  vshd    wk4, wk5, wk4
                  stws,ma wk4, 4(dest)

                  ldw     16(buf0), wk4
                  ldw     20(buf0), wk3
                  ldw     24(buf0), wk2
                  ldw     28(buf0), wk1

                  vshd    wk5, wk4, wk5
                  stws,ma wk5, 4(dest)
                  vshd    wk4, wk3, wk4
                  stws,ma wk4, 4(dest)
                  vshd    wk3, wk2, wk3
                  stws,ma wk3, 4(dest)
                  vshd    wk2, wk1, wk2
                  stws,ma wk2, 4(dest)

                  sub,<=  ntr, 0, 0       ; if (ntr > 0) bring other buffer
                  ldwx    chunk(buf0), 0  ;   into cache.

;               swap current and "other" staging buffer.

                  add       buf0, chunk, buf0   ; buf0 -> next line
                  sub       0, chunk, chunk

vr_n_venom_entry
                comb,>=,n ntw, 0, vr_n_venom_loop
                ldo     -32(ntw), ntw
;             }
                ldo 32(ntw), ntw

;           |- buf0 -> staging buffer
;           |- ntw < 32
;           |- ntr == nr of bytes left in venom_read < 32
;           |- ntr < 0 => wk1 contains last few bytes of copy span
;           |- wk1 contains eventual m.s. end of *dest, aligned as source.
;           |- 0 < ntw - ntr < 4

;           src= buf0;  // set src to point to remaining data in staging buffer.
              copy buf0, src

;         }
;     }

n_tail
;   while (ntr > 0)
      b,n  n_tail_entry
;     {
;       wk2= *(src++); ntr-= 4;
n_tail_loop
          ldo       -4(ntr), ntr
;       wk1= [wk1.wk2]>> shiftb;
          vshd      wk1, wk2, wk1
;       if (ntw <= 4) goto suffix;
          comib,>=,n  4, ntw, suffix
;       *(dest++)= wk1; ntw-= 4;
          ldo       -4(ntw), ntw
          stws,ma   wk1, 4(0,dest)
;       wk2= wk1;
          copy      wk2, wk1
n_tail_entry
          comb,>,n  ntr, 0, n_tail_loop
          ldws,ma   4(0,src), wk2
;     }

n_fixup

;   |- dest -> last word affected by move
;   |- 0 <= ntw < 4 && ntw == nr of bytes to copy to last word of dest
;   |- wk1 == ntw bytes of *dest, source aligned

;   wk1= [wk1.0] >> shiftb;
      vshd  wk1, 0, wk1

; }

; SUFFIX SECTION ==========================================

suffix
; |- 0 <= ntw <= 4
; |- ntw == 4 => wk1 == [SSSSSSSS]
; |- 0 < ntw < 4 => wk1 == [SSSSxxxx] (ntw bytes of SS)
; |- ntw == 0 => do nothing
; |- wk1 == ntw bytes of *dest, destination aligned

; if (ntw)
    comb,=,n  0, ntw, suffix_done
;   {

;     if (ntw != 4)
        comib,=  4, ntw, suffix_write
        copy wk1, wk2
;       {

;         wk2= *dest /* but not in checksum space */;
            ldil    l%0xf3800000, chunk
            copy    dest, dest_nock
            and     dest, chunk, wk2
            sub,<>  chunk, wk2, 0
            dep     0, 8, 1, dest_nock

            ldws  0(0,dest_nock), wk2

;         wk2= bytes from src | bytes from dest
            zdep  ntw, 28, 29, chunk    ; chunk= 8*ntw - 1
            ldo   -1(chunk), chunk      ;
            mtsar chunk
            vextru  wk1, 32, wk1
            vdep    wk1, 32, wk2
;       }
suffix_write

          stws,ma wk2, 4(0,dest)
;   }
suffix_done

; RETURN ==================================================

return
#ifdef RETURN_PTRS
	copy	src, ret0
	copy    dest, ret1
#else
; return (cksum_unit) ? *cksum_unit : 0
        ldw     cksum_unit(sp), wk1
        copy    0, ret0
        extru,= wk1, 31, 32, 0
        ldw     (wk1), ret0
#endif
        .leave

; END =====================================================


;* SETUP_VENOM
; short subroutine to set up venom for a copy.
; Call using  "bl   setup_venom,link".
; Call uses registers as follows:
;   chunk = number of words for venom to copy - 1
;   src   = address of first word of source
;   venom = nothing special
;   link  = return address
; On return,
;   chunk unchanged
;   src   unchanged
;   link  unchanged
;   venom corrupted

setup_venom
        ldil  l%venom_user_base, venom
        ldo   r%venom_user_base(venom), venom
        stws  0, venom_user_ctrl(venom)       ; 0 == block move command.
        stws  chunk, venom_span_count(venom)
        bv    0(link)
        stws  src, venom_graphics_addr(venom)

        .procend

; ============================================================================
;*
;*                  Design and Implementation Notes
;*                  -------------------------------
;* 
;* Med_copy and med_vcopy have much code in common.  In much of what follows
;* references to med_copy apply equally to med_vcopy.
;*
;* A call to med_copy proceeds in three stages, which we refer to as the
;* Prefix, the Copy, and the Suffix.  Before discussing the stages in detail,
;* here are med_copy's major variables:
;* 
;*     delta_s:  the number of bytes which precede the first byte of the
;*          source area (pointed to by 'src') in the word which contains it.
;*          0 <= delta_s < 4.
;* 
;*     delta_d:  the number of bytes which precede the first byte of the
;*          destination are (pointed to by 'dest') in the word which
;*          contains it.  0 <= delta_d < 4.
;* 
;*     shiftb:  the number of bits the contents of a source word must be shifted
;*          "right" (to the least significant end of the word) to align it
;*          for writing to the destination area. (4+delta_d-delta_s)%4 * 8.
;* 
;*     src, dest:  must both be corrected to be word-aligned, effectively by
;*          subtracting delta_s and delta_d from the appropriate initial
;*          values.
;* 
;*     ntw: initially, the number of bytes to copy, becoming the number of
;*          bytes to WRITE to the destination.
;* 
;*     ntr: the number of bytes to be READ from the source.
;* 
;*     wkN: working registers (1 <= N <= 5) hold data on the move.  Wk1
;*          conveys source data from one phase of the med_copy to the next.
;*          Wk2 holds data from the destination when the prefix and suffix
;*          are being assembled.  Wk3, wk4, and wk5 are used only during
;*          the copy loops (the registers they name have other uses at other
;*          times).
;*
;* Comments
;* --------
;*
;* The code contains several kinds of comments.  Most of the comments in the
;* body are a pidgin-C version of what follows in the program.  Extracting
;* those lines with, say, "grep '^;' med_copy.s" gives you a detailed
;* overview of the program.
;*
;* Some comments start with `|-'.  These are low-grade assertions which
;* attempt to give an idea of the state at that point.
;*
;* Finally, there's the stuff you're reading at the moment.  Try printing
;* a separate copy of this and having it handy as you venture into the
;* murky depths of the code.
;*
;* 
;* Organisation
;* ------------
;*
;* The code is structured as follows:
;*
;*     Entry points:   med_copy
;*                     med_vcopy
;*
;*     Parameter analysis:  set control flags according to the source and
;*         destination addresses supplied.
;*
;*     Prepare checksum unit, if required:  If either source or destination
;*	   is in checksummed VRAM, initialise the checksum unit with a value
;*	   that compensates for the bytes at either end of the copy range
;*	   which med_copy reads but does not copy (cf, the prefix and suffix
;*	   described later).
;*
;*     Prefix processing:  two sections, for aligned and non-aligned copies.
;*         (non-aligned are those where med_copy must shift bytes left or
;*         right as it copies them).
;*
;*     The copy routines:  two sections, aligned and non-aligned.  Each
;*         section contains several copy routines:  Venom moves; memory
;*         to IO; IO to memory, and the `last few words'.
;*
;*     Suffix processing:  finishing off the last few bytes of the copy.
;*
;*     Result building:  retrieve and return the checksum, etc.
;*
;* Control flow is mostly top-to-bottom; it looks far worse than it really is.
;* 
;*
;* Analysing Parameters
;* --------------------
;* 
;* This section examines the src and dest parameters in turn and sets 
;* control flags (in register `flags') as follows:
;*
;*   io_read:   src is in IO space (ie, > 0xf0000000).
;*   io_write:  dest is in IO space (ie, > 0xf0000000).
;*   do_cksum:  either src or dest is in the checksummed area of a medusa
;*              card's address space.
;*   use_venom: src is in IO space, dest is not, and function was called as
;*              med_vcopy.
;*
;* In addition, if do_cksum is set, this section sets `wk1' to the address
;* (either src or dest) which is in checksummed RAM.  It also sets `dest_nock'
;* to the non-checksumming alternative to dest, for use in the prefix section.
;*
;* NOTE:  If both src and dest are in checksummed RAM, the destination takes
;* priority and src is altered to be non-checksumming.  [The reasoning is
;* that it's hard to imagine anyone doing this, but if they were, it would
;* be because they were moving data from an old packet into a new one, so
;* only the checksum in the new area is of interest.]
;*
;* Preparing the Checksum Unit
;* ---------------------------
;*
;* This section calculates the initial value for the on-card checksum unit.
;* Normally this is zero.  However, med_copy is often asked to move data
;* which do not fill an integral number of words (eg, len%4 != 0) or which
;* do not fall with word boundaries (eg, src%4 != 0, (src+len)%4 != 0, etc).
;* In these cases, med_copy forms the initial value as the complement of the
;* one's complement sum of the `waste' portions of the words at either end
;* of the copy range.
;*
;* The Prefix
;* ----------
;* 
;* This stage examines med_copy's parameters, selects which copy algorithm
;* to use, and prepares to execute it.
;* 
;* When delta_d != 0, med_copy has to preserve the most significant bytes of
;* the word containing the first byte of the destination area.  These bytes
;* form the "prefix" which gives this stage its name.  The prefix is one or two
;* words which, when shifted by the appropriate amount by the copy, produce the
;* desired first word of the destination.  It is formed by extracting portions
;* of the first words of both source and destination.
;* 
;* There are three major cases to consider, depending on the relative values of
;* delta_s and delta_d:
;* 
;*     delta_d == delta_s:  The source and destination areas are "aligned"
;*          w.r.t each other, and the copy does not have to shift the data
;*          while copying it.  Therefore, when at least four bytes are being
;*          copied, the prefix is written to *dest before the main copy
;*          begins.  Example, delta_s = delta_d = 2:
;* 
;*                 +--------+--------+--------+--------+
;*          src    |....0011|22334455| + + +
;*                 +--------+--------+--------+--------+
;*                      vvvv
;*                 +--------+
;*          prefix |AABB0011|
;*                 +--------+
;*                  ^^^^
;*                 +--------+--------+--------+--------+
;*          dest   |AABB....|........| + + +
;*                 +--------+--------+--------+--------+
;* 
;*     delta_d < delta_s:  Data must be shifted "left" as it moves from the
;*          source to the destination.  Thus the prefix must be shifted
;*          RIGHT when planted at the head of the source.  Example:
;*          delta_d = 1, delta_s = 3.
;* 
;*                 +--------+--------+--------+--------+
;*          src    |......00|11223344|55667788| + + +
;*                 +--------+--------+--------+--------+
;*                        vv vvvv
;*                      +--------+
;*          prefix      |AA001122|
;*                      +--------+
;*                  //---^
;*                 +--------+--------+--------+--------+
;*          dest   |AA......|........| + + +
;*                 +--------+--------+--------+--------+
;* 
;*     delta_s < delta_d:  Data must be shifted "right" as it moves from the
;*          source to the destination.  Thus the prefix must be shifted LEFT
;*          when planted at the head of the source.  This requires TWO
;*          registers to hold the doctored source.  When at least four bytes
;*          are being copied, the prefix code itself shifts and writes the
;*          first word of the prefix.  Example, delta_d = 3, delta_s = 1:
;* 
;*                          +--------+--------+--------+
;*          src             |..001122|33445566| + + +
;*                          +--------+--------+--------+
;*                             vvvvvv
;*                 +--------+--------+
;*          prefix |....AABB|CC001122|
;*                 +--------+--------+
;*                      ^----\\\\\\
;*                          +--------+--------+--------+
;*          dest            |AABBCC..|........| + + +
;*                          +--------+--------+--------+
;* 
;* When delta_d == 0, the prefix is simply the first word of the source.  This
;* gives the same initial conditions for the copy phase as the other cases.
;* 
;* 
;* The Copy
;* --------
;* 
;* The copy phase is essentially simple:  read some words from the source,
;* write them to the destination, and repeat.  Complications arise for several
;* reasons:
;*     
;*     (mis)alignment of src and dest:  when delta_s != delta_d, the
;*          function must shift the data as it copies it:  
;*
;*                      +--------+--------+--------+
;*                      |00010203|04050607|08091011|
;*                      +--------+--------+--------+
;*                       v___
;*                           v  
;*                      +--------+--------+--------+
;*                      |....0001|02030405|06070809|
;*                      +--------+--------+--------+
;*
;*          In effect, each word of the source contains data for two words
;*          in the destination.  A copy loop has to preserve the last word
;*          read during one pass for use at the start of the next.  All
;*          loops in med_copy and med_vcopy expect the register wk1 to
;*          contain this word.  The PREFIX code described earlier supplies
;*          the right value in wk1 for the first word to be copied.
;* 
;*     (ntw % 4) != 0:  in effect, some bytes in the last word of the
;*          destination must be preserved.  This means the copies must
;*          finish without writing the last word of the range, and pass
;*          it instead to the suffix phase.
;* 
;*     performance:  moving several words at a time gives much better
;*          throughput than doing one at a time.  Thus the copy proceeds
;*          first in large chunks, then finishes the last few words
;*          individually.
;* 
;*          The size of a chunk depends on whether the copy is to or from
;*          IO space.  If writing to IO space from memory, the optimum size
;*          is two words; if reading from IO space, the optimum is four.
;* 
;* Dealing with these variations, especially performance, means that the
;* copy section actually contains EIGHT main copy routines:
;*
;*      pr_a_copy - Programmed Read (IO -> store), source and dest'n Aligned
;*      pw_a_copy - Programmed Write(store -> IO), source and dest'n aligned
;*      p_a_tail  - Programmed Read/Write, aligned - last few words
;*      vr_a_copy - Venom Read, aligned
;*
;*      pr_n_copy - Programmed Read, source and dest'n not aligned
;*      pw_n_copy - Programmed Write, source and dest'n not aligned
;*      p_n_tail  - Programmed Read/Write, not aligned, last few words
;*      vr_n_copy - Venom Read, not aligned
;*
;* `Programmed' means using ordinary load and store instructions; `venom'
;* means using some special block move instructions provided by the graphics
;* function unit in the 700 series' processor-memory controller (see later).
;*
;* All expect src and dest to point to the next words to be read and written,
;* respectively, and ntw to contain the number of bytes to write.  The non-
;* aligned copies also expect ntr to contain the number of bytes to read,
;* shift to contain the number of bits to shift data to the `right', and wk1
;* to contain the prefix, ie, the eventual most significant bytes of the word
;* at *dest.  These conditions also hold at the end of each copy routine.
;* At the very end, ntr <= 0, 0 <= ntw <= 4, and wk1 contains the remaining
;* bytes to be deposited at *dest.
;*
;*
;* Using Venom
;* -----------
;* 
;* While programmed writes -- ordinary Store instructions -- give adequate
;* performance (38 MB/s), programmed reads are considerably slower (10 MB/s).
;* This is partly because locations in IO space are not cached, and partly
;* because reading is a two-way transaction whereas writing is one-way
;* (`issue address, wait, read data' vs.  `issue address, issue data').
;*
;* The 700 series has a unit called Venom which offers a number of functions
;* intended to improve graphics performance.  One function is a block move,
;* which copies a block of at most 4096 words between IO and CPU memory.
;*
;* A Venom move has two stages.  The CPU prepares Venom for the move by
;* writing a command, the `span count' (number of words to move, less one),
;* and the address of the first word in IO space, into control registers.
;* Then it moves successive chunks of data using a variation of the Flush
;* Data Cache instruction:  Graphics Flush Write moves data from IO to
;* memory, and GF Read moves data from memory to IO.  (Note that the names
;* have the opposite sense to what you might expect.)
;*
;* A GF Write or Read instruction has an address A as its operand, and moves
;* up to eight words, until it exhausts the span count or reaches the end of
;* the cache line addressed by A.  The exact number of words is given by
;* 
;*               min((span count + 1), (8 - (A % 32)/4))
;*
;* In practice, GF Read gives little improvement over programmed writes for
;* sending data to IO; in certain common cases, it can be worse.  GF Write,
;* on the other hand, is nearly twice as fast for moving data from IO to 
;* memory as programmed reads.  Med_vcopy, therefore, uses venom to read
;* data into memory from IO.  Aligned and non-aligned copies have quite
;* different implementations.
;*
;* Med_vcopy implements aligned copies directly from IO to the memory address
;* given by `dest'.  The copy takes up to three stages.  The first deals with
;* the case where dest points to a word in the middle of a cache line (ie,
;* dest%32 != 0), so that the gfw moves fewer than eight words.  The second
;* stage moves all remaining whole words in the range of the copy (in cache-
;* aligned chunks of eight words).  The final stage deals with the case where
;* the last word of the source contains one to three bytes of data to copy;
;* it loads that word into a register and passes it on to the suffix code.
;* The total number of words moved by venom is (ntr/4).
;*
;* Non-aligned copies use two `staging buffers' in med_vcopy's stack frame.
;* Each corresponds to one full cache line.  Copies proceed by filling one
;* staging buffer with the next chunk of data from the source while shifting
;* and copying the contents of the other to the destination.  The overlap
;* between filling one buffer and emptying the other is possible because
;* while the gfw instruction is loading memory, the shift-and-copy is
;* operating mostly within the data cache.  The code also uses `cache hint'
;* instructions (load register 0) to ensure the buffer just loaded is in
;* the cache ready for the following shift-and-copy.  The remainder of the
;* processing is similar to the corresponding programmed copy.  The total
;* number of words moved by venom is ((ntr+3)/4), possibly one greater than
;* for the aligned case, because it's OK to read the (possibly partial) final
;* word into the staging buffer.
;* 
;* 
;* The Suffix
;* ----------
;* 
;* The "suffix" is those bytes in the last word of the destination that are
;* not in the range of the copy.  On entry to this section, wk1 holds the
;* last bytes of the source (correctly aligned for writing), and ntw holds
;* the number of bytes to copy to the destination (0 to 4).
;* 
;* 
;* Performance
;* -----------
;* 
;* Measured on a 9000/720 (50 MHz clock), med_copy typically gives these results
;* for copy throughput in various cases (cycles/byte, MBytes/s):
;* 
;*                Cache -> cache  Cache -> IO  IO -> Cache  IO -> Cache
;*                                             (programmed)   (venom)
;*   Aligned        1.00  50.0    1.31  38.1   4.88  10.2   2.70  18.5
;*   Non-aligned    1.06  47.1    1.43  35.0   5.00  10.0   2.70  18.5
;* 
;* Overhead, ie, time not spent copying, is comparable to that for memcpy
;* (36 to 72 cycles per call) for the programmed cases.

#endif
